// Copyright 2025 Zentrum für Digitale Souveränität der Öffentlichen Verwaltung (ZenDiS) GmbH.
// SPDX-License-Identifier: MIT

import { FunctionComponent, useMemo } from 'react'
import {
    Bar,
    BarChart,
    CartesianGrid,
    ReferenceLine,
    ResponsiveContainer,
    XAxis,
    YAxis,
} from 'recharts'
import { CheckResult } from '../lib/utils'
import { Badge } from './ui/badge'
import {
    Card,
    CardContent,
    CardDescription,
    CardFooter,
    CardHeader,
    CardTitle,
} from './ui/card'
import { ChartContainer, ChartTooltip, ChartTooltipContent } from './ui/chart'

interface Stats {
    chartData: Array<{ value: number; count: number }>
    nullCount: number
    domain: [number, number]
    mean: number
    stddev: number
    outliers: number
}
interface Props {
    title: string
    description: string
    propertyKey: keyof CheckResult
    results: Array<CheckResult>

    xLabel: string
    yLabel: string

    // for highly skewed dataset it might be useful to even show outliers.
    // maybe thats actually the only data we have.
    disableOutlierDetection?: boolean
    currentScanRepositoryURL?: string
    currentScan?: {
        properties: {
            value: number
        }
    }
}

const filterOutliers = (someArray: number[]): number[] => {
    if (someArray.length < 4) return someArray

    let values, q1, q3, iqr, maxValue, minValue

    values = someArray.slice().sort((a, b) => a - b) //copy array fast and sort

    if ((values.length / 4) % 1 === 0) {
        //find quartiles
        q1 =
            (1 / 2) *
            (values[values.length / 4] + values[values.length / 4 + 1])
        q3 =
            (1 / 2) *
            (values[values.length * (3 / 4)] +
                values[values.length * (3 / 4) + 1])
    } else {
        q1 = values[Math.floor(values.length / 4 + 1)]
        q3 = values[Math.ceil(values.length * (3 / 4) + 1)]
    }

    iqr = q3 - q1
    maxValue = q3 + iqr * 1.5
    minValue = q1 - iqr * 1.5

    return [minValue, maxValue]
}

const Hist: FunctionComponent<Props> = ({
    results,
    title,
    description,
    disableOutlierDetection,
    currentScan,
    xLabel,
    yLabel,
    propertyKey,
    currentScanRepositoryURL,
}) => {
    const stats: Stats = useMemo(() => {
        // remove all without any issueReactionTime
        const cleaned = results.filter((result) => result[propertyKey] !== null)

        let minValueBarrier = Math.min(
            ...cleaned.map((result) => result[propertyKey] as number),
        )
        let maxValueBarrier = Math.max(
            ...cleaned.map((result) => result[propertyKey] as number),
        )

        if (!disableOutlierDetection) {
            const [min, max] = filterOutliers(
                cleaned.map((result) => result[propertyKey] as number),
            )

            minValueBarrier = min
            maxValueBarrier = max
        }

        let outliers = 0
        // now count the occurrence of each issueReactionTime
        const counted = cleaned.reduce(
            (acc, result) => {
                if (
                    (result[propertyKey] as number) < minValueBarrier ||
                    (result[propertyKey] as number) > maxValueBarrier
                ) {
                    outliers++
                    return acc
                }

                if (acc[result[propertyKey] as number]) {
                    acc[result[propertyKey] as number]++
                } else {
                    acc[result[propertyKey] as number] = 1
                }
                return acc
            },
            {} as Record<number, number>,
        )

        // get max and min
        const max = Math.max(
            ...Object.keys(counted).map((propertyKey) => parseInt(propertyKey)),
        )
        const min = Math.min(
            ...Object.keys(counted).map((propertyKey) => parseInt(propertyKey)),
        )

        // calculate the best bins amount using Sturges' formula
        const bins = Math.ceil(1 + Math.log2(cleaned.length))
        const binSize = (max - min) / bins
        // iterate over the data and count the occurrences in each bin
        const binsMap = Array.from({ length: bins }, (_, i) => {
            const start = min + i * binSize
            const end = min + (i + 1) * binSize
            return {
                value: `${start.toFixed(0)}-${end.toFixed(0)}`,
                count: Object.entries(counted).reduce(
                    (acc, [value, count]) =>
                        acc +
                        (parseInt(value) >= start && parseInt(value) < end
                            ? count
                            : 0),
                    0,
                ),
            }
        }).reduce(
            (acc, { value, count }) => {
                acc[value] = count
                return acc
            },
            {} as Record<string, number>,
        )

        // calculate mean days (values)
        const mean =
            Object.entries(counted).reduce(
                (acc, [value, count]) => acc + parseInt(value) * count,
                0,
            ) / cleaned.length

        // calculate stddev
        const stddev = Math.sqrt(
            Object.entries(counted).reduce(
                (acc, [value, count]) =>
                    acc + count * Math.pow(parseInt(value) - mean, 2),
                0,
            ) / cleaned.length,
        )

        return {
            chartData: Object.entries(binsMap).map(([value, count]) => ({
                value: parseInt(value),
                count,
            })),
            nullCount: results.length - cleaned.length,
            domain: [0, Math.max(...Object.values(counted))],
            mean,
            stddev,
            outliers,
        }
    }, [results, propertyKey, disableOutlierDetection])

    return (
        <Card>
            <CardHeader>
                <CardTitle>{title}</CardTitle>
                <CardDescription>{description}</CardDescription>
            </CardHeader>
            <CardContent className="relative">
                <ResponsiveContainer width={'100%'} height={300}>
                    <ChartContainer
                        config={{
                            value: {
                                label: xLabel,
                            },
                            count: {
                                label: yLabel,
                            },
                        }}
                    >
                        {stats ? (
                            <BarChart
                                margin={{ top: 20 }}
                                width={730}
                                height={250}
                                data={stats.chartData}
                            >
                                <YAxis
                                    fontSize={14}
                                    domain={stats.domain}
                                    allowDecimals={false}
                                />
                                <CartesianGrid vertical={false} />
                                <XAxis
                                    dataKey="value"
                                    tickSize={4}
                                    minTickGap={50}
                                    fontSize={14}
                                    tickMargin={8}
                                />
                                <ChartTooltip
                                    content={
                                        <ChartTooltipContent
                                            labelFormatter={(
                                                label,
                                                [payload],
                                            ) => {
                                                return `${xLabel}: ${payload.payload.value} < X`
                                            }}
                                            indicator="dot"
                                        />
                                    }
                                />
                                <Bar
                                    dataKey="count"
                                    stroke="var(--chart-2)"
                                    fill="var(--chart-2)"
                                    type="step"
                                    fillOpacity={1}
                                />
                                {currentScan && (
                                    <ReferenceLine
                                        stroke="hsl(var(--primary))"
                                        strokeWidth={2}
                                        x={currentScan.properties.value}
                                    ></ReferenceLine>
                                )}
                            </BarChart>
                        ) : (
                            <div>Loading...</div>
                        )}
                    </ChartContainer>
                </ResponsiveContainer>
                {currentScan && (
                    <span className="text-[14px] text-primary">
                        {currentScanRepositoryURL}:{' '}
                        {currentScan.properties.value !== null ? (
                            <Badge>{currentScan?.properties.value}</Badge>
                        ) : (
                            <Badge>No Value</Badge>
                        )}
                    </span>
                )}
                <span className="text-muted-foreground absolute left-0 top-1/3 -rotate-90 text-sm">
                    {yLabel}
                </span>
                <span className="text-muted-foreground absolute bottom-2 left-1/2 -translate-x-20 text-sm">
                    {xLabel}
                </span>
            </CardContent>
            <CardFooter>
                <div className="flex w-full items-start gap-2 text-sm">
                    <div className="grid gap-2">
                        <div className="items-center gap-2 font-medium leading-none">
                            <div>
                                Projects with a value{' '}
                                <Badge variant={'outline'}>
                                    {stats?.chartData.reduce(
                                        (acc, data) => acc + data.count,
                                        0,
                                    )}
                                </Badge>
                            </div>

                            <div>
                                Amount of projects without a value{' '}
                                <Badge variant={'outline'}>
                                    {stats?.nullCount}
                                </Badge>
                            </div>
                        </div>
                        <div className="text-muted-foreground flex items-center gap-2 leading-none">
                            Mean: {stats?.mean.toFixed(2)}, Std:{' '}
                            {stats?.stddev.toFixed(2)}
                            {stats.outliers > 0 &&
                                `, Outliers:
                            ${stats?.outliers}`}
                        </div>
                    </div>
                </div>
            </CardFooter>
        </Card>
    )
}

export default Hist
